---
title: "PyPI (Python)"
order: 13
---

The Python Package Index ([PyPI](https://pypi.org/)) is the main repository for software and libraries written in the Python programming language. Since version 5.2, ProGet supports the creation of PyPI feeds to host and deploy custom Python software packages.

ProGet's ["Feeds"](/docs/proget/feeds/feed-overview) can act as PyPI package repositories that let you centralize your package management and keep a handle on which packages your organization uses. Along with [publishing and hosting your private Python packages](/docs/proget/feeds/pypi/howto-pypi-publish), ProGet also acts as a [proxy for the PyPI repository](/docs/proget/feeds/pypi/howto-pypi-proxy), allowing you to cache and manage dependencies pulled from [PyPI](https://pypi.org/).

## Prerequisites

In order to create, publish and install Python packages from ProGet, you need [Python](https://www.python.org/downloads/) installed. You can confirm installation with `python --version`.

The documentation on this page also requires that [pip](https://pip.pypa.io/) be installed (Confirm installation with `pip --version`), however you can also use [Pipenv](https://pipenv.pypa.io/en/latest/) or [Poetry](https://python-poetry.org/):

### Using PipEnv

To use Pipenv for managing your packages, you will need to install it using `pip install`:

```bash
$ pip install pipenv
```

Confirm installation with `pipenv --version`.

### Using Poetry

To use Poetry, you will need to install via the official installer:

```bash
$ curl -sSL https://install.python-poetry.org | python3 -
```

Confirm installation with `poetry --version`.

## Using Your PyPI Feed as a Source to Install Packages

To use your PyPI feed when installing packages you can either set it globally in your with the `pip config` command, or run the `pip install` command. 

### Using a `pip Config` File

Setting your feed globally in the [pip config](https://pip.pypa.io/en/stable/topics/configuration/) file using the [`pip config`](https://pip.pypa.io/en/stable/cli/pip_config/) command makes it a default source for all installations. To store your feed you will need to use `pip config` with a `--global` parameter containing your feed endpoint URL.

```bash
$ pip config --global set global.index-url https://«proget-server»/pypi/«feed-name»/simple 
```

This command will generate a `pip config` file that looks like:

```bash
[global]
index-url = https://«proget-server»/pypi/«feed-name»/simple 
```

Then you can install packages simply using the `pip install` command:

```bash
pip install «package-name»==«package-version»
```

::: (Info) (Note: Scoping)
The `pip config` can be scoped to global (`--global`), user (`--user`), and to the environment (`--site`). The commands above are scoped to the global scope.
:::

### Using `pip install` With a URL Parameter

For one-off installations, you can also use `pip install`. Note this is not persistent —meaning you’ll need to re-specify the URL each time you install from your feed. Using a [pip config](https://pip.pypa.io/en/stable/topics/configuration/) file is more convenient for long-term use.

To install Python packages with the URL in the `pip install` command, you will need to add a `--extra-index-url` parameter containing your feed endpoint URL:

```bash
$ pip install «package-name»==«package-version» --extra-index-url https://«proget-server»/pypi/«feed-name»/simple
```

### Using Pipenv

To configure [Pipenv](https://pipenv.pypa.io/en/latest/) to use your PyPI feed, set the `PIP_INDEX_URL` environment variable in your shell before running `pipenv install`, or include it in your Pipfile:

```bash
$ export PIP_INDEX_URL=https://«proget-server»/pypi/«feed-name»/simple
```

Alternatively, modify your Pipfile to include:

```bash
[[source]]
name = "«feed-name»"
url = "https://«proget-server»/pypi/«feed-name»/simple"
verify_ssl = true
```

Then install packages with:

```bash
$ pipenv install «package-name»==«package-version»
```

### Using Poetry

To configure [Poetry](https://python-poetry.org/) to use your PyPI feed, you can add the repository to your project using the `poetry config` command:

```bash
$ poetry config repositories.«feed-name» https://«proget-server»/pypi/«feed-name»/simple
```

Then, to install a package from your feed, use the `poetry add` command:

```bash
poetry add «package-name» --source «feed-name»
```

## Authenticating to a PyPI Feed

If you've configured your feed to require authentication, you can use `pip` to authenticate to it, or alternatively use [PipEnv](#authenticate-pipenv) or [Poetry](#authenticate-poetry). 

:::(info) (Authenticating with an API Key)
Feeds can be authenticated with a ProGet username and password (`«username»:«password»`), but we strongly recommend using an [API Key](/docs/proget/reference-api/proget-apikeys) for authentication, with `api` as the username, and then API Key as the password (`api:«api-key»`)
:::

### Authenticating with `pip config`

Using `pip config` securely stores credentials in your [pip config](https://pip.pypa.io/en/stable/topics/configuration/) file, protecting your credentials. This is also convenient for repeated installs, removing the need to re-enter credentials.

To use a [pip config](https://pip.pypa.io/en/stable/topics/configuration/) file to store the authenticated feed, run the `pip config` command:

```bash
$ pip config --global set global.index-url https://api:«api-key»@«proget-server»/pypi/«feed-name»/simple 
$ pip config --global set global.trusted-host «proget-server»
```

These commands will generate a `pip config` file that looks like:

```bash
[global]
index-url = https://api:«api-key»@«proget-server»/pypi/«feed-name»/simple 
trusted-host = «proget-server»
```

Then you can install packages simply using the `pip install` command:

```bash
pip install «package-name»==«package-version»
```

### Authenticating with `pip install`

Alternatively you can authenticate with `pip install`. This is good for quick, one-off installations but exposes sensitive information in plain text, both in your shell history and process list. For this reason storing your credentials in your pip config file is generally recommended.

To pass basic authentication along with the URL in the `pip install` command, you will need to add a `--trusted-host` parameter containing your ProGet host and insert the API Key into the beginning of your URL. The `pip install` command would look like this:

```bash
$ pip install «package-name»==«package-version» --trusted-host «proget-server» -i https://api:«api-key»@«proget-server»/pypi/«feed-name»/simple 
```

Using `pip install` is A more secure alternative is to use `pip config`.

### Authenticating with Pipenv { #authenticate-pipenv }

To use your feed with Pipenv, you can set the `PIP_INDEX_URL` environment variable with the authenticated URL:

```bash
$ export PIP_INDEX_URL=https://api:«api-key»@«proget-server»/pypi/«feed-name»/simple
```

Then, install packages with Pipenv:

```bash
$ pipenv install «package-name»==«package-version»
```

### Authenticating with Poetry { #authenticate-poetry }

To configure Poetry to authenticate with your PyPI feed, add the repository and credentials to your Poetry configuration:

```bash
$ poetry config repositories.pypy-feed https://«proget-server»/pypi/«feed-name»/simple
$ poetry config http-basic.pypy-feed api «api-key»
```

Once configured, you can install packages using:

```bash
$ poetry add «package-name» --source pypy-feed
```

## Creating Python Packages

To create a Python package that can be hosted by ProGet, you will need `setuptools` installed by running:

```bash
$ pip install setuptools wheel
```

You will then need to run the standard `python` command:

```bash
$ python setup.py sdist bdist_wheel
```

For more detail on how to create a Python package, read our [HOWTO: Publish npm Packages to a Private Registry in ProGet](/docs/proget/feeds/pypi/howto-pypi-publish#create-package)

## Uploading Python Packages to a PyPI Feed

To publish your Python package to your ProGet PyPI feed, you can use [pgutil](/docs/proget/reference-api/proget-pgutil).

pgutil will require some [minor configuration](/docs/proget/reference-api/proget-pgutil#sources) before use. This includes setting up your ProGet instance and API key as a source by running:

```bash
$ pgutil sources add --name=Default --url=«proget-url» --api-key=«api-key»
```

To upload your package you would enter:

```bash
$ pgutil packages upload --feed=«feed-name» --input-file=«path-to-package»
```

For more examples on uploading, read [HOWTO: Publish npm Packages to a Private Registry in ProGet](/docs/proget/feeds/pypi/howto-pypi-publish#upload-package)

### Uploading with Twine

It's also possible to use the `twine upload` command if want to use [twine](https://pypi.org/project/twine/). To use an [API Key](/docs/proget/reference-api/proget-apikeys), you will need to use `api` for the username and the API key as the password. 

:::(warn)(Feed Endpoint URL)
Note that the feed endpoint URLs used when pushing packages vary slightly from the standard feed URL, ending with `/legacy` rather than `/simple`
:::

### Using .pypirc to Upload Python Packages

You can also set your PyPI feed globally to publish packages to by editing the `.pypirc` file located in `~\Users\«user»\` as follows:

```bash
[distutils] 
index-servers = proget 

[proget] 
repository = http://«proget-server»/pypi/«feed-name»/legacy
username = «username» 
password = «password»
```

For a more detailed guide on publishing your own Python packages you can read [HOWTO: Upload Python Packages to a Private PyPI Repository in ProGet](/docs/proget/feeds/pypi/howto-pypi-publish#upload-package)

## Technical Limitations

### ProGet 2024

#### Connectors
Most PyPI repositories do not provide a user interface and are not designed to be searched or browsed, which makes their display in ProGet challenging. Specifically:

Package Searching

 * PyPI repositories do not support searching.  However, you can still navigate to a specific package page by typing the exact name in the search box.
 * When possible, ProGet displays a default list of packages in a feed sorted by download count. For PyPI connectors, ProGet instead uses the RSS feed which is provided by PyPI.org to expose the latest package uploads.

Non-PyPI.org repositories

 * PyPI.org implements a JSON API that provides some extra metadata for packages (such as a readme file), but most other repositories do not have this
 * Very basic information can be browsed in the ProGet UI if you know the exact name of the package, but this should primarily be used as a proxy instead of browsed directly
 * You may need to use the "Advanced" tab of the connector to specify an alternative `/simple` path; for example, if you try to connect to `https://download.pytorch.org/whl/cu124` without specifying this, you will receive `403` errors; changing the simple path to `/` will resolve this

### ProGet 2023
 * ProGet only supports searching package name and summary fields via the XML-RPC API (implemented by PyPI.org), other use cases of this API are not supported by ProGet
 * Connectors to other PyPI feeds require the PyPI JSON API to be implemented (both ProGet and PyPI.org support this)
 * Package names are generally normalized (as per the [PEP503 specs](https://www.python.org/dev/peps/pep-0503)), but packages where versions only differ by `-` and `_` (e.g. `my-package 1.0` and `my_package 1.1`) aren't properly listed under `my-package`. In addition, UI-based searching does not treat `_` and `-` the same
 
If these limitations are more than minor inconveniences, [please let us know](https://forums.inedo.com/) so we can figure out how to fix them.
